import abc
import itertools
import pathlib
import torch
from .extensions import model_configs, merge_methods, merge_spaces
from .extensions.merge_spaces import MergeSpace
from typing import Optional, Dict, Tuple, Union


class RecipeNode(abc.ABC):
    @abc.abstractmethod
    def accept(self, visitor, *args, **kwargs):
        pass

    @property
    @abc.abstractmethod
    def merge_space(self) -> merge_spaces.MergeSpace:
        pass

    @property
    @abc.abstractmethod
    def model_config(self) -> Optional[model_configs.ModelConfig]:
        pass

    @abc.abstractmethod
    def __contains__(self, item):
        pass

    def set_cache(self, _cache: dict = ...):
        return self

    def __add__(self, other: "RecipeNodeOrValue") -> "RecipeNode":
        other = merge_methods.value_to_node(other)
        base, delta = self, other
        if other.merge_space == "weight":
            base, delta = other, self
        return merge_methods.resolve("add_difference")(base, delta)

    def __radd__(self, other: "RecipeNodeOrValue") -> "RecipeNode":
        other = merge_methods.value_to_node(other)
        return other + self

    def __sub__(self, other: "RecipeNodeOrValue") -> "RecipeNode":
        other = merge_methods.value_to_node(other)
        return merge_methods.resolve("subtract")(self, other)

    def __rsub__(self, other: "RecipeNodeOrValue") -> "RecipeNode":
        other = merge_methods.value_to_node(other)
        return other - self

    def __or__(self, other: "RecipeNodeOrValue") -> "RecipeNode":
        other = merge_methods.value_to_node(other)
        return merge_methods.resolve("fallback")(self, other)

    def __ror__(self, other: "RecipeNodeOrValue") -> "RecipeNode":
        other = merge_methods.value_to_node(other)
        return other | self

    def to(self, *, device: Optional[Union[str, torch.device, "RecipeNode"]] = None, dtype: Optional[Union[str, torch.dtype, "RecipeNode"]] = None):
        if isinstance(device, torch.device):
            device = str(device)
        if isinstance(dtype, torch.dtype):
            from sd_mecha.extensions.builtin.merge_methods import cast_dtype_map_reversed
            dtype = cast_dtype_map_reversed[dtype]
        return merge_methods.resolve("cast")(self, device=device, dtype=dtype)


PythonLiteralValue = str | int | float | bool | type(None)
NonDictLiteralValue = PythonLiteralValue | torch.Tensor
LiteralValue = NonDictLiteralValue | dict
RecipeNodeOrValue = RecipeNode | LiteralValue | pathlib.Path


class LiteralRecipeNode(RecipeNode):
    def __init__(
        self,
        value: LiteralValue,
        *,
        model_config: Optional[str | model_configs.ModelConfig] = None,
        merge_space: Optional[str | merge_spaces.MergeSpace] = None,
    ):
        self.value = value
        self.__model_config = model_configs.resolve(model_config) if isinstance(model_config, str) else model_config
        self.__merge_space = merge_spaces.resolve(merge_space) if isinstance(merge_space, str) else merge_space
        if isinstance(self.value, dict):
            first_value = next(iter((*self.value.values(), Ellipsis)))  # using ... as placeholder as None is a valid LiteralValue
            if isinstance(first_value, RecipeNode):
                if model_config is None:
                    self.__model_config = first_value.model_config
                if merge_space is None:
                    self.__merge_space = first_value.merge_space
                if not all(v.model_config == self.__model_config for v in self.value.values()):
                    raise ValueError(f"All model configs should be the same, expected {self.__model_config} but got {set(v.model_config for v in self.value.values())}")
                if not all(v.merge_space == first_value.merge_space for v in self.value.values()):
                    raise ValueError(f"All merge spaces should be the same, but got multiple: {set(v.merge_space for v in self.value.values())}")
        if self.__merge_space is None:
            self.__merge_space = merge_spaces.resolve("param")

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_literal(self, *args, **kwargs)

    @property
    def merge_space(self) -> MergeSpace:
        return self.__merge_space

    @property
    def model_config(self) -> Optional[model_configs.ModelConfig]:
        return self.__model_config

    @model_config.setter
    def model_config(self, model_config: Optional[model_configs.ModelConfig]):
        self.__model_config = model_config

    def __contains__(self, item):
        if isinstance(item, LiteralRecipeNode):
            return self.value == item.value
        else:
            return False


class ModelRecipeNode(RecipeNode):
    def __init__(
        self,
        path: pathlib.Path,
        *,
        model_config: Optional[str | model_configs.ModelConfig] = None,
        merge_space: str | MergeSpace = "weight",
    ):
        if not isinstance(path, pathlib.Path):
            raise TypeError(f"The type of 'state_dict' must be Path, not {type(path).__name__}")

        self.path = path
        self.state_dict = None
        self.__model_config = model_configs.resolve(model_config) if isinstance(model_config, str) else model_config
        self.__merge_space = merge_spaces.resolve(merge_space) if isinstance(merge_space, str) else merge_space

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_model(self, *args, **kwargs)

    @property
    def merge_space(self) -> merge_spaces.MergeSpace:
        return self.__merge_space

    @property
    def model_config(self) -> Optional[model_configs.ModelConfig]:
        return self.__model_config

    @model_config.setter
    def model_config(self, value: Optional[model_configs.ModelConfig]):
        self.__model_config = value

    def __contains__(self, item):
        if isinstance(item, ModelRecipeNode):
            return self.path == item.path
        else:
            return False


class MergeRecipeNode(RecipeNode):
    def __init__(
        self,
        merge_method,
        args: Tuple[RecipeNode, ...],
        kwargs: Dict[str, RecipeNode],
        cache: dict = None,
    ):
        self.merge_method = merge_method
        self.args = args
        self.kwargs = kwargs
        self.cache = cache
        self.__validate_args()

    def __validate_args(self):
        if not isinstance(self.merge_space, MergeSpace):
            raise RuntimeError(f"Could not infer merge space from arguments for method {self.merge_method.identifier}")

    def accept(self, visitor, *args, **kwargs):
        return visitor.visit_merge(self, *args, **kwargs)

    @property
    def merge_space(self) -> merge_spaces.MergeSpace:
        return self.merge_method.get_return_merge_space(
            [v.merge_space for v in self.args],
            {k: v.merge_space for k, v in self.kwargs.items()},
        )

    @property
    def model_config(self) -> Optional[model_configs.ModelConfig]:
        return self.merge_method.get_return_config(
            [v.model_config for v in self.args],
            {k: v.model_config for k, v in self.kwargs.items()},
        )

    def __contains__(self, item):
        return self is item or any(
            item in v
            for v in itertools.chain(self.args, self.kwargs.values())
            if isinstance(v, RecipeNode)
        )

    def set_cache(self, cache: dict = ...):
        if cache is Ellipsis:
            cache = {}

        self.cache = cache
        return self


class RecipeVisitor(abc.ABC):
    @abc.abstractmethod
    def visit_literal(self, node: LiteralRecipeNode):
        pass

    @abc.abstractmethod
    def visit_model(self, node: ModelRecipeNode):
        pass

    @abc.abstractmethod
    def visit_merge(self, node: MergeRecipeNode):
        pass


class ModelDepthRecipeVisitor(RecipeVisitor):
    def visit_literal(self, node: LiteralRecipeNode):
        return 0

    def visit_model(self, _node: ModelRecipeNode):
        return 1

    def visit_merge(self, node: MergeRecipeNode):
        return max(
            child.accept(self)
            for children in (node.args, node.kwargs.values())
            for child in children
        ) + 1


class ModelsCountVisitor(RecipeVisitor):
    def __init__(self):
        self.__seen_nodes = []

    def visit_literal(self, node: LiteralRecipeNode) -> int:
        return 0

    def visit_model(self, node: ModelRecipeNode) -> int:
        seen = node in self.__seen_nodes
        self.__seen_nodes.append(node)
        return int(not seen)

    def visit_merge(self, node: MergeRecipeNode) -> int:
        return sum(
            child.accept(self)
            for children in (node.args, node.kwargs.values())
            for child in children
        )



import torch
from torch import Tensor
from sd_mecha import merge_method, StateDict, Return, Parameter
from sd_mecha.extensions import model_configs
from .convert_huggingface_sd_vae_to_original import convert_vae


sdxl_kohya_config = model_configs.resolve('sdxl-kohya')
sdxl_sgm_config = model_configs.resolve('sdxl-sgm')


@merge_method(
    identifier=f"convert_'{sdxl_kohya_config.identifier}'_to_'{sdxl_sgm_config.identifier}'",
    is_conversion=True,
)
def convert_sdxl_kohya_to_original(
    kohya_sd: Parameter(StateDict[Tensor], model_config=sdxl_kohya_config),
    **kwargs,
) -> Return(Tensor, model_config=sdxl_sgm_config):
    sgm_key = kwargs["key"]

    if sgm_key.startswith("model.diffusion_model."):
        kohya_key = sgm_key.replace("model.diffusion_model.", "unet.")
        return kohya_sd[kohya_key]
    elif sgm_key.startswith("conditioner.embedders.0."):
        kohya_key = sgm_key.replace("conditioner.embedders.0.transformer.", "te1.")
        return kohya_sd[kohya_key]
    elif sgm_key.startswith("conditioner.embedders.1."):
        if sgm_key.endswith("text_projection"):
            kohya_key = "te2.text_projection.weight"
        else:
            kohya_key = sgm_key.replace("conditioner.embedders.1.model.", "te2.text_model.")
            kohya_key = kohya_key.replace(".token_embedding.", ".embeddings.token_embedding.")
            kohya_key = kohya_key.replace(".positional_embedding", ".embeddings.position_embedding.weight")
            kohya_key = kohya_key.replace(".transformer.resblocks.", ".encoder.layers.")
            kohya_key = kohya_key.replace(".attn.", ".self_attn.")
            kohya_key = kohya_key.replace(".mlp.c_fc.", ".mlp.fc1.")
            kohya_key = kohya_key.replace(".mlp.c_proj.", ".mlp.fc2.")
            kohya_key = kohya_key.replace(".ln_final.", ".final_layer_norm.")
            kohya_key = kohya_key.replace(".ln_", ".layer_norm")

        if kohya_key.endswith((".in_proj_weight", ".in_proj_bias")):
            is_bias = kohya_key.endswith("bias")
            partial_key = kohya_key.replace(".in_proj_weight", "").replace(".in_proj_bias", "")
            res = torch.vstack([
                kohya_sd[f"{partial_key}.{k}_proj.{'bias' if is_bias else 'weight'}"]
                for k in ("q", "k", "v")
            ])
        else:
            res = kohya_sd[kohya_key]

        if sgm_key.endswith("text_projection"):
            res = res.T

        return res
    elif sgm_key.startswith("first_stage_model."):
        return convert_vae(kohya_sd, sgm_key)
    else:
        return kohya_sd[sgm_key]



import torch
from sd_mecha.extensions.merge_methods import StateDict


def convert_vae(huggingface_sd: StateDict[torch.Tensor], ldm_key: str) -> torch.Tensor:
    huggingface_key = ".".join(ldm_key.split(".")[1:])

    needs_reshape = False
    for sd_weight_name, fake_weight_name in vae_extra_conversion_map.items():
        if f"mid.attn_1.{sd_weight_name}.weight" in huggingface_key or f"mid.attn_1.{sd_weight_name}.bias" in huggingface_key:
            needs_reshape = True
            huggingface_key = huggingface_key.replace(sd_weight_name, fake_weight_name)

    for weight_name in vae_extra_conversion_map.values():
        if f"mid.attn_1.{weight_name}.weight" in huggingface_key:
            needs_reshape = True

    if "attentions" in huggingface_key:
        for sd_part, hf_part in vae_conversion_map_attn.items():
            huggingface_key = huggingface_key.replace(sd_part, hf_part)

    for sd_part, hf_part in vae_conversion_map.items():
        huggingface_key = huggingface_key.replace(sd_part, hf_part)

    huggingface_key = f"vae.{huggingface_key}"
    res = huggingface_sd[huggingface_key]
    if needs_reshape:
        res = reshape_weight_for_sd(res)
    return res


def reshape_weight_for_sd(w):
    # convert HF linear weights to SD conv2d weights
    if w.ndim != 1:
        return w.reshape(*w.shape, 1, 1)
    else:
        return w


vae_conversion_map = {
    # (original, huggingface)
    "nin_shortcut": "conv_shortcut",
    "norm_out": "conv_norm_out",
    "mid.attn_1.": "mid_block.attentions.0.",
}
for i in range(4):
    # down_blocks have two resnets
    for j in range(2):
        hf_down_prefix = f"encoder.down_blocks.{i}.resnets.{j}."
        sd_down_prefix = f"encoder.down.{i}.block.{j}."
        vae_conversion_map[sd_down_prefix] = hf_down_prefix

    if i < 3:
        hf_downsample_prefix = f"down_blocks.{i}.downsamplers.0."
        sd_downsample_prefix = f"down.{i}.downsample."
        vae_conversion_map[sd_downsample_prefix] = hf_downsample_prefix

        hf_upsample_prefix = f"up_blocks.{i}.upsamplers.0."
        sd_upsample_prefix = f"up.{3-i}.upsample."
        vae_conversion_map[sd_upsample_prefix] = hf_upsample_prefix

    # up_blocks have three resnets
    # also, up blocks in hf are numbered in reverse from sd
    for j in range(3):
        hf_up_prefix = f"decoder.up_blocks.{i}.resnets.{j}."
        sd_up_prefix = f"decoder.up.{3-i}.block.{j}."
        vae_conversion_map[sd_up_prefix] = hf_up_prefix

# this part accounts for mid blocks in both the encoder and the decoder
for i in range(2):
    hf_mid_res_prefix = f"mid_block.resnets.{i}."
    sd_mid_res_prefix = f"mid.block_{i+1}."
    vae_conversion_map[sd_mid_res_prefix] = hf_mid_res_prefix


vae_conversion_map_attn = {
    # (original, huggingface)
    "norm.": "group_norm.",
    "q.": "query.",
    "k.": "key.",
    "v.": "value.",
    "proj_out.": "proj_attn.",
}


# This is probably not the most ideal solution, but it does work.
vae_extra_conversion_map = {
    # (original, huggingface)
    "q": "to_q",
    "k": "to_k",
    "v": "to_v",
    "proj_out": "to_out.0",
}



import abc
import typing
from types import UnionType
from typing import runtime_checkable, Protocol, TypeVar


K = TypeVar("K")
V = TypeVar("V")


@runtime_checkable
class WriteOnlyMapping(Protocol[K, V]):
    @abc.abstractmethod
    def __setitem__(self, key: K, value: V) -> None:
        ...

    @abc.abstractmethod
    def __len__(self) -> int:
        ...


def is_subclass(source: type | UnionType, target: type | UnionType):
    source_origin = typing.get_origin(source) or source
    target_origin = typing.get_origin(target) or target
    if isinstance(source_origin, TypeVar):
        return False
    if isinstance(target_origin, TypeVar):
        return any(is_subclass(source_origin, constraint) for constraint in target_origin.__constraints__)
    if issubclass(source_origin, UnionType):
        return all(is_subclass(arg, target) for arg in typing.get_args(source))
    return issubclass(source_origin, target)



from torch import Tensor
from sd_mecha import merge_method, Parameter, Return, StateDict, StateDictKeyError


@merge_method(is_conversion=True)
def convert_flux_to_backbone_only(
    flux: Parameter(StateDict[Tensor], model_config="flux-flux"),
    **kwargs,
) -> Return(Tensor, model_config="flux-flux_diffuser_only"):
    diffuser_only_key = kwargs["key"]
    full_key = f"model.diffusion_model.{diffuser_only_key}"
    return flux[full_key]


@merge_method(is_conversion=True)
def convert_flux_backbone_to_full(
    flux_diffuser: Parameter(StateDict[Tensor], model_config="flux-flux_diffuser_only"),
    **kwargs,
) -> Return(Tensor, model_config="flux-flux"):
    full_key = kwargs["key"]
    if not full_key.startswith("model.diffusion_model."):
        raise StateDictKeyError(full_key)

    diffuser_only_key = full_key[len("model.diffusion_model."):]
    return flux_diffuser[diffuser_only_key]



import abc
import functools
import inspect
import pathlib
import fuzzywuzzy.process
import torch
from typing import Mapping, Protocol, Optional, runtime_checkable, List
from sd_mecha.streaming import InSafetensorsDict, OutSafetensorsDict
from sd_mecha.typing_ import WriteOnlyMapping


@runtime_checkable
class ModelFormat(Protocol):
    identifier: str

    @abc.abstractmethod
    def matches(self, path: pathlib.Path) -> bool:
        ...

    @abc.abstractmethod
    def get_read_dict(self, path: pathlib.Path, buffer_size: int) -> Mapping[str, torch.Tensor]:
        ...

    @abc.abstractmethod
    def get_write_dict(
        self,
        path: pathlib.Path,
        model_config,
        mecha_recipe: str,
        buffer_size: int,
    ) -> WriteOnlyMapping[str, torch.Tensor]:
        ...


def register_model_format(
    model_format: Optional[type(ModelFormat)] = None, *,
    identifier: Optional[str] = None,
):
    if model_format is None:
        return lambda model_format: __register_model_format_impl(model_format, identifier=identifier)
    return __register_model_format_impl(model_format, identifier=identifier)


def __register_model_format_impl(
    model_format: type(ModelFormat), *,
    identifier: Optional[str],
):
    if not inspect.isclass(model_format):
        raise ValueError(f"model_format must be a class, not {type(ModelFormat)}")

    if identifier is None:
        identifier = model_format.__name__

    if identifier in _model_format_registry:
        raise ValueError(f"model format '{identifier}' already exists")

    model_format = model_format()
    model_format.identifier = identifier
    _model_format_registry[identifier] = model_format


def resolve(identifier: str) -> ModelFormat:
    try:
        return _model_format_registry[identifier]
    except KeyError as e:
        suggestion = fuzzywuzzy.process.extractOne(str(e), _model_format_registry.keys())[0]
        raise ValueError(f"unknown model format: {e}. Nearest match is '{suggestion}'")


def get_all() -> List[ModelFormat]:
    return list(_model_format_registry.values())


_model_format_registry = {}


@functools.cache
def _register_builtin_model_formats():
    @register_model_format(identifier="single_file")
    class SingleFileModelFormat(ModelFormat):
        def matches(self, path: pathlib.Path) -> bool:
            path = path.resolve()
            return path.suffix == ".safetensors"

        def get_read_dict(self, path: pathlib.Path, buffer_size: int) -> Mapping[str, torch.Tensor]:
            return InSafetensorsDict(path, buffer_size)

        def get_write_dict(
            self,
            path: pathlib.Path,
            model_config,
            mecha_recipe: str,
            buffer_size: int,
        ) -> WriteOnlyMapping[str, torch.Tensor]:
            return OutSafetensorsDict(path, model_config.metadata, mecha_recipe, buffer_size)


_register_builtin_model_formats()



from sd_mecha import merge_method, Parameter, Return, StateDict
from torch import Tensor


@merge_method
def exchange_ema(
    model: Parameter(StateDict[Tensor]),
    **kwargs,
) -> Return(Tensor):
    input_keys = model.model_config.keys()
    target_key = kwargs["key"]
    to_ema_key_fn = to_ema_key_fns.get(model.model_config.identifier, lambda k: k)
    ema_key = to_ema_key_fn(target_key)

    if ema_key in input_keys:
        return model[ema_key]
    else:
        for input_key in input_keys:
            if to_ema_key_fn(input_key) == target_key:
                return model[input_key]
        return model[target_key]


to_ema_key_fns = {
    "sd1-ldm": lambda k: f"model_ema.{k[len('model.'):].replace('.', '')}"
}



import logging
import pathlib
from typing import List, Optional, Hashable
from .extensions import merge_methods
from .recipe_nodes import RecipeNode, ModelRecipeNode, RecipeVisitor, LiteralRecipeNode, MergeRecipeNode


MECHA_FORMAT_VERSION = "0.1.0"


def deserialize_path(recipe: pathlib.Path) -> RecipeNode:
    if not recipe.exists():
        raise ValueError(f"unable to deserialize '{recipe}': no such file")

    if recipe.suffix == ".mecha":
        with open(recipe, "r") as recipe_file:
            return deserialize(recipe_file.read())
    else:
        raise ValueError(f"unable to deserialize '{recipe}': unknown extension")


def deserialize(recipe: str | List[str]) -> RecipeNode:
    """
    Recreate a recipe graph from its serialized `.mecha` format.

    Args:
        recipe (str or List[str]):
            The textual representation (as a string or list of lines) of the recipe.

    Returns:
        A `RecipeNode` that can be further merged or manipulated.
    """
    if not isinstance(recipe, list):
        recipe = recipe.split("\n")

    if not recipe[0].startswith("version"):
        raise RuntimeError("bad format: expected version at line 1")

    actual_version, recipe = recipe[0], recipe[1:]
    expected_version = get_version_header(MECHA_FORMAT_VERSION)
    if actual_version != expected_version:
        raise RuntimeError(f"bad recipe version: got {actual_version}, expected {expected_version}")

    results = []

    def parse(line):
        line = line.strip()
        if line.startswith("#"):
            return

        command, *args = tokenize(line)
        positional_args, keyword_args = preprocess_args(args)
        if command == "dict":
            results.append(dict(*positional_args, **keyword_args))
        elif command == "literal":
            results.append(LiteralRecipeNode(*positional_args, **keyword_args))
        elif command == "model":
            path = pathlib.Path(positional_args[0])
            results.append(ModelRecipeNode(path, *positional_args[1:], **keyword_args))
        elif command == "merge":
            method, *positional_args = positional_args
            method = merge_methods.resolve(method)
            results.append(method(*positional_args, **keyword_args))
        else:
            raise ValueError(f"unknown command: {command}")

    def preprocess_args(args):
        positional_args = []
        named_args = {}
        for arg_index, arg in enumerate(args):
            if '=' in arg:
                key, value = arg.split('=', maxsplit=1)
                named_args[key] = get_arg_value(value, arg_index)
            else:
                positional_args.append(get_arg_value(arg, arg_index))
        return positional_args, named_args

    def get_arg_value(arg, arg_index):
        try:
            if arg in CONSTANTS:
                return CONSTANTS[arg]
            elif arg.startswith('&'):
                ref_index = int(arg[1:])
                if ref_index < 0 or ref_index >= len(results):
                    raise ValueError(f"reference {arg} out of bounds")
                return results[ref_index]
            elif arg.startswith('"') and arg.endswith('"'):
                return arg[1:-1]
            elif '.' in arg or 'e' in arg.lower():
                return float(arg)
            else:
                return int(arg)
        except ValueError as e:
            raise ValueError(f"argument {arg_index}: {str(e)}")

    def tokenize(line):
        tokens = []
        current_token = []
        quote_prefix = []
        inside_quotes = False
        is_escape = False
        for char in line:
            if is_escape:
                is_escape = False
            elif char == "\\":
                is_escape = True
                continue
            elif char == '"':
                inside_quotes = not inside_quotes
                if inside_quotes:  # Begin of quoted string
                    quote_prefix = current_token
                    current_token = []
                else:  # End of quoted string
                    tokens.append(f'{"".join(quote_prefix)}"{"".join(current_token)}"')
                    current_token = []
                    quote_prefix = []
                continue
            elif char == ' ' and not inside_quotes:
                if current_token:  # End of a token
                    tokens.append(''.join(current_token))
                    current_token = []
                continue
            current_token.append(char)
        if inside_quotes:  # Handle mismatched quotes
            raise ValueError(f"mismatched quotes in input")
        if current_token:  # Add last token if exists
            tokens.append(''.join(current_token))
        return tokens

    for line_num, line in enumerate(recipe, 1):
        try:
            parse(line)
        except ValueError as e:
            raise ValueError(f"line {line_num}: {e}.\n    {line}")

    return results[-1]


def serialize(recipe: RecipeNode, *, output: Optional[pathlib.Path | str] = None) -> str:
    """
    Convert a recipe graph into a string that captures its merge instructions.

    This is the first step of persisting a recipe to disk in `.mecha` format.

    Args:
        recipe:
            A `RecipeNode` describing the merge.
        output:
            Path to the output file to save.

    Returns:
        A string representation of the recipe, suitable for writing to a .mecha file.
    """
    serializer = SerializerVisitor()
    recipe.accept(serializer)
    version_header = get_version_header(MECHA_FORMAT_VERSION)
    serialized = "\n".join([version_header] + serializer.instructions)

    if isinstance(output, str):
        output = pathlib.Path(output)
    if output is not None:
        output = output.absolute()
        logging.info(f"Saving recipe to {output}")
        output.write_text(serialized)

    return serialized


def get_version_header(version: str):
    return f"version {version}"


class SerializerVisitor(RecipeVisitor):
    def __init__(self, instructions: Optional[List[str]] = None):
        self.instructions = instructions if instructions is not None else []

    def visit_literal(self, node: LiteralRecipeNode):
        value = self.__serialize_value(node.value)
        if node.model_config is None:
            return value
        else:
            config = self.__serialize_value(node.model_config.identifier)
            merge_space = self.__serialize_value(node.merge_space.identifier)
            line = f"literal {value} model_config={config} merge_space={merge_space}"
            return self.__add_instruction(line)

    def visit_model(self, node: ModelRecipeNode) -> str:
        path = self.__serialize_value(str(node.path))
        config = self.__serialize_value(getattr(node.model_config, "identifier", None))
        merge_space = self.__serialize_value(node.merge_space.identifier)
        line = f"model {path} model_config={config} merge_space={merge_space}"
        return self.__add_instruction(line)

    def visit_merge(self, node: MergeRecipeNode) -> str:
        identifier = self.__serialize_value(node.merge_method.get_identifier())
        parts = ["merge", identifier] + [
            self.__serialize_value(v)
            for v in node.args
        ] + [
            f"{k}={self.__serialize_value(v)}"
            for k, v in node.kwargs.items()
        ]
        line = " ".join(parts)
        return self.__add_instruction(line)

    def __serialize_value(self, value) -> str:
        if isinstance(value, str):
            value = value.replace("\\", "\\\\").replace('"', "\\\"")
            return f'"{value}"'
        if isinstance(value, dict):
            dict_line = "dict " + " ".join(f"{k}={self.__serialize_value(v)}" for k, v in value.items())
            return self.__add_instruction(dict_line)
        if isinstance(value, (int, float)):
            return str(value)
        # int or float needs to be handled before this (1.0 == True)
        if isinstance(value, Hashable) and value in REVERSE_CONSTANTS:
            return REVERSE_CONSTANTS[value]
        if isinstance(value, RecipeNode):
            return value.accept(self)
        raise TypeError(f"type {type(value)} cannot be serialized: {value}")

    def __add_instruction(self, instruction: str) -> str:
        try:
            return f"&{self.instructions.index(instruction)}"
        except ValueError:
            self.instructions.append(instruction)
            return f"&{len(self.instructions) - 1}"


CONSTANTS = {
    "null": None,
    "true": True,
    "false": False,
}
REVERSE_CONSTANTS = {
    v: k for k, v in CONSTANTS.items()
}
